from plone.registry.field import DisallowedProperty
from plone.registry.field import InterfaceConstrainedProperty
from plone.registry.field import is_primitive
from plone.registry.field import StubbornProperty
from plone.registry.interfaces import IPersistentField
from zope.component import adapter
from zope.interface import implementer
from zope.schema.interfaces import IChoice
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.interfaces import IField
from zope.schema.interfaces import ISource
from zope.schema.vocabulary import SimpleVocabulary

import plone.registry.field


@implementer(IPersistentField)
@adapter(IField)
def persistentFieldAdapter(context):
    """Turn a non-persistent field into a persistent one"""

    if IPersistentField.providedBy(context):
        return context

    # See if we have an equivalently-named field

    class_name = context.__class__.__name__
    persistent_class = getattr(plone.registry.field, class_name, None)
    if persistent_class is None:
        return None

    if not issubclass(persistent_class, context.__class__):
        __traceback_info__ = "Can only clone a field of an equivalent type."
        return None

    ignored = list(DisallowedProperty.uses + StubbornProperty.uses)
    constrained = list(InterfaceConstrainedProperty.uses)

    instance = persistent_class.__new__(persistent_class)

    context_dict = {k: v for k, v in context.__dict__.items() if k not in ignored}

    for key, iface in constrained:
        value = context_dict.get(key, None)
        if value is None or value == context.missing_value:
            continue
        value = iface(value, None)
        if value is None:
            __traceback_info__ = (
                "The property `{}` cannot be adapted to "
                "`{}`.".format(
                    key,
                    iface.__identifier__,
                )
            )
            return None
        context_dict[key] = value

    instance.__dict__.update(context_dict)
    return instance


@implementer(IPersistentField)
@adapter(IChoice)
def choicePersistentFieldAdapter(context):
    """Special handling for Choice fields."""
    instance = persistentFieldAdapter(context)
    if instance is None:
        return None

    if ISource.providedBy(context.vocabulary) or IContextSourceBinder.providedBy(
        context.vocabulary
    ):
        safe = False

        # Attempt to reverse engineer a 'values' argument
        if isinstance(context.vocabulary, SimpleVocabulary):
            values = []
            safe = True
            for term in context.vocabulary:
                if term.token == str(term.value) and is_primitive(term.value):
                    values.append(term.value)
                else:
                    safe = False
                    break
            if safe:
                instance._values = values

        if not safe:
            __traceback_info__ = (
                "Persistent fields only support named vocabularies "
                "or vocabularies based on simple value sets."
            )
            return None

    return instance



from persistent import Persistent
from plone.registry.events import RecordModifiedEvent
from plone.registry.interfaces import IInterfaceAwareRecord
from plone.registry.interfaces import IPersistentField
from plone.registry.interfaces import IRecord
from zope.dottedname.resolve import resolve
from zope.event import notify
from zope.interface import alsoProvides
from zope.interface import implementer


_marker = object()


@implementer(IRecord)
class Record(Persistent):
    """A record that is stored in the registry.

    If __parent__ is set, consider this a "bound" record. In this case, the
    field and value are read from and written to the parent registry.

    BBB: The current storage implementation does not actually store Record
    objects directly. However, we keep the Persistent base class so that old
    values may be loaded during automated migration.
    """

    __name__ = ""
    __parent__ = None

    def __init__(self, field, value=_marker, _validate=True):
        if _validate and not IPersistentField.providedBy(field):
            raise ValueError("Field is not persistent")

        if value is _marker:
            value = field.default
        else:
            if _validate:
                if value != field.missing_value:
                    bound_field = field.bind(self)
                    bound_field.validate(value)

        # Bypass event notification and setting on the parent
        self._field = field
        self._value = value

        if field.interfaceName:
            alsoProvides(self, IInterfaceAwareRecord)

    # The 'field' property

    def _get_field(self):
        if self.__parent__ is not None:
            return self.__parent__.records._getField(self.__name__)
        return self._field

    def _set_field(self, value):
        if self.__parent__ is not None:
            self.__parent__.records._setField(self.__name__, value)
        self._field = value

    _field = None
    field = property(_get_field, _set_field)

    # The 'value' property

    def _get_value(self):
        if self.__parent__ is not None:
            return self.__parent__.records._values[self.__name__]
        return self._value

    def _set_value(self, value):
        field = self.field

        if field is None:
            raise ValueError("The record's field must be set before its value")

        field = field.bind(self)
        if value != field.missing_value:
            field.validate(value)

        oldValue = self._value
        self._value = value

        if self.__parent__ is not None:
            self.__parent__.records._values[self.__name__] = value

        notify(RecordModifiedEvent(self, oldValue, value))

    _value = None
    value = property(_get_value, _set_value)

    # Interface name, field name and interface instance

    @property
    def interfaceName(self):
        return self.field.interfaceName

    @property
    def fieldName(self):
        return self.field.fieldName

    @property
    def interface(self):
        try:
            return resolve(self.interfaceName)
        except ImportError:
            return None

    # Print representation

    def __repr__(self):
        return "<Record %s>" % self.__name__



"""This module defines persistent versions of various fields.

The idea is that when a record is created, we copy relevant field properties
from the transient schema field from zope.schema, into the corresponding
persistent field. Not all field types are supported, but the common types
are.
"""
from persistent import Persistent
from plone.registry.interfaces import IPersistentField
from zope.interface import implementer
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.vocabulary import SimpleVocabulary

import zope.schema
import zope.schema._field


try:
    import plone.schema

    HASPLONESCHEMA = True
except ImportError:
    HASPLONESCHEMA = False

_primitives = (int, bool, str, bytes, tuple, list, set, frozenset, dict, float)

_missing_value_marker = object()


def is_primitive(value):
    return value is None or isinstance(value, _primitives)


class DisallowedProperty:
    """A property that may not be set on an instance. It may still be set
    defined in a base class.
    """

    uses = []

    def __init__(self, name):
        self._name = name
        DisallowedProperty.uses.append(name)

    def __get__(self, inst, type_=None):
        # look for the object in bases
        if type_ is not None:
            for c in type_.__mro__:
                if self._name in c.__dict__ and not isinstance(
                    c.__dict__[self._name], DisallowedProperty
                ):
                    function = c.__dict__[self._name]
                    return function.__get__(inst, type_)
        raise AttributeError(self._name)

    def __set__(self, inst, value):
        raise ValueError(
            "Persistent fields does not support setting the `{}` "
            "property".format(self._name)
        )


class StubbornProperty:
    """A property that stays stubbornly at a single, pre-defined value."""

    uses = []

    def __init__(self, name, value):
        StubbornProperty.uses.append(name)
        self._name = name
        self._value = value

    def __set__(self, inst, value):
        pass

    def __get__(self, inst, type_=None):
        return self._value


class InterfaceConstrainedProperty:
    """A property that may only contain values providing a certain interface."""

    uses = []

    def __init__(self, name, interface):
        InterfaceConstrainedProperty.uses.append((name, interface))
        self._name = name
        self._interface = interface

    def __set__(self, inst, value):
        if value != inst.missing_value and not self._interface.providedBy(value):
            raise ValueError(
                "The property `{}` may only contain objects "
                "providing `{}`.".format(
                    self._name,
                    self._interface.__identifier__,
                )
            )
        inst.__dict__[self._name] = value


@implementer(IPersistentField)
class PersistentField(Persistent):
    """Base class for persistent field definitions."""

    # Persistent fields do not have an order
    order = StubbornProperty("order", -1)

    # We don't allow setting a custom constraint, as this would introduce a
    # dependency on a symbol such as a function that may go away
    constraint = DisallowedProperty("constraint")

    # Details about which interface/field name we originally came form, if any
    interfaceName = None
    fieldName = None


class PersistentCollectionField(PersistentField, zope.schema._field.AbstractCollection):
    """Ensure that value_type is a persistent field"""

    value_type = InterfaceConstrainedProperty("value_type", IPersistentField)


class Bytes(PersistentField, zope.schema.Bytes):
    pass


class BytesLine(PersistentField, zope.schema.BytesLine):
    pass


class ASCII(PersistentField, zope.schema.ASCII):
    pass


class ASCIILine(PersistentField, zope.schema.ASCIILine):
    pass


class Text(PersistentField, zope.schema.Text):
    pass


class TextLine(PersistentField, zope.schema.TextLine):
    pass


class Bool(PersistentField, zope.schema.Bool):
    pass


class Int(PersistentField, zope.schema.Int):
    pass


class Float(PersistentField, zope.schema.Float):
    pass


class Decimal(PersistentField, zope.schema.Decimal):
    pass


class Tuple(PersistentCollectionField, zope.schema.Tuple):
    pass


class List(PersistentCollectionField, zope.schema.List):
    pass


class Set(PersistentCollectionField, zope.schema.Set):
    pass


class FrozenSet(PersistentCollectionField, zope.schema.FrozenSet):
    pass


class Password(PersistentField, zope.schema.Password):
    pass


class Dict(PersistentField, zope.schema.Dict):
    key_type = InterfaceConstrainedProperty("key_type", IPersistentField)
    value_type = InterfaceConstrainedProperty("value_type", IPersistentField)


class Datetime(PersistentField, zope.schema.Datetime):
    pass


class Date(PersistentField, zope.schema.Date):
    pass


class Timedelta(PersistentField, zope.schema.Timedelta):
    pass


class SourceText(PersistentField, zope.schema.SourceText):
    pass


class URI(PersistentField, zope.schema.URI):
    pass


class Id(PersistentField, zope.schema.Id):
    pass


class DottedName(PersistentField, zope.schema.DottedName):
    pass


class Choice(PersistentField, zope.schema.Choice):
    # We can only support string name or primitive=list vocabularies
    _values = None
    _vocabulary = None

    def __init__(self, values=None, vocabulary=None, source=None, **kw):
        if vocabulary is not None and not isinstance(vocabulary, str):
            values = self._normalized_values(vocabulary)
            if values is None:
                raise ValueError(
                    "Persistent fields only support named vocabularies or "
                    "vocabularies based on simple value sets."
                )
            vocabulary = None
        elif source is not None:
            raise ValueError(
                "Persistent fields do not support sources, only named "
                "vocabularies or vocabularies based on simple value sets."
            )

        assert not (
            values is None and vocabulary is None
        ), "You must specify either values or vocabulary."
        assert (
            values is None or vocabulary is None
        ), "You cannot specify both values and vocabulary."

        self.vocabularyName = None

        if values is not None:
            for value in values:
                if not is_primitive(value):
                    raise ValueError(
                        "Vocabulary values may only contain primitive values."
                    )
            self._values = values
        else:
            self.vocabularyName = vocabulary

        self._init_field = bool(self.vocabularyName)
        super(zope.schema.Choice, self).__init__(**kw)
        self._init_field = False

    def _normalized_values(self, vocabulary):
        if getattr(vocabulary, "__iter__", None):
            if all([isinstance(term.value, str) for term in vocabulary]):
                return [term.value for term in vocabulary]
        return None

    @property
    def vocabulary(self):
        # may be set by bind()
        if self._vocabulary is not None:
            return self._vocabulary
        if self._values is not None:
            return SimpleVocabulary.fromValues(self._values)

    DisallowedProperty.uses.append("vocabulary")

    # override bind to allow us to keep constraints on the 'vocabulary'
    # property
    def bind(self, object):
        clone = zope.schema.Field.bind(self, object)
        # get registered vocabulary if needed:
        if IContextSourceBinder.providedBy(self.vocabulary):
            clone._vocabulary = self.vocabulary(object)
            assert zope.schema.interfaces.ISource.providedBy(clone.vocabulary)
        elif clone.vocabulary is None and self.vocabularyName is not None:
            vr = zope.schema.vocabulary.getVocabularyRegistry()
            clone._vocabulary = vr.get(object, self.vocabularyName)
            assert zope.schema.interfaces.ISource.providedBy(clone.vocabulary)
        return clone


if HASPLONESCHEMA:

    class JSONField(PersistentField, plone.schema.JSONField):
        key_type = InterfaceConstrainedProperty("key_type", IPersistentField)
        value_type = InterfaceConstrainedProperty("value_type", IPersistentField)



from plone.registry.interfaces import IFieldRef
from zope.interface import implementedBy
from zope.interface import implementer


@implementer(IFieldRef)
class FieldRef:
    """Default field reference."""

    def __init__(self, name, originalField):
        self.recordName = name
        self.originalField = originalField

    @property
    def __providedBy__(self):
        provided = getattr(self, "__provides__", None)
        if provided is None:
            provided = implementedBy(self.__class__)

        return provided + self.originalField.__providedBy__

    def __getattr__(self, name):
        return getattr(self.originalField, name)



from plone.registry.fieldfactory import choicePersistentFieldAdapter
from plone.registry.fieldfactory import persistentFieldAdapter
from zope import schema
from zope.component import eventtesting
from zope.component import provideAdapter
from zope.component import testing
from zope.interface import Interface

import doctest
import re
import unittest


IGNORE_U = doctest.register_optionflag("IGNORE_U")


class PolyglotOutputChecker(doctest.OutputChecker):
    def check_output(self, want, got, optionflags):
        # fix changed objectfield class in zope4
        got = re.sub(
            "zope.schema._field.Object", "zope.schema._bootstrapfields.Object", got
        )

        if hasattr(self, "_toAscii"):
            got = self._toAscii(got)
            want = self._toAscii(want)

        # Naive fix for comparing byte strings
        if got != want and optionflags & IGNORE_U:
            got = re.sub(r'^u([\'"])', r"\1", got)
            want = re.sub(r'^u([\'"])', r"\1", want)

        return doctest.OutputChecker.check_output(self, want, got, optionflags)


class IMailSettings(Interface):
    """Settings for email"""

    sender = schema.TextLine(title="Mail sender", default="root@localhost")
    smtp_host = schema.URI(title="SMTP host server")


class IMailPreferences(Interface):
    """Settings for email"""

    max_daily = schema.Int(title="Maximum number of emails per day", min=0, default=3)
    settings = schema.Object(title="Mail settings to use", schema=IMailSettings)


def setUp(test=None):
    testing.setUp()
    eventtesting.setUp()

    provideAdapter(persistentFieldAdapter)
    provideAdapter(choicePersistentFieldAdapter)


class TestBugs(unittest.TestCase):
    """Regression tests for bugs that have been fixed"""

    def setUp(self):
        setUp(self)

    def tearDown(self):
        testing.tearDown(self)

    def test_bind_choice(self):
        from plone.registry.field import Choice
        from zope.schema.vocabulary import getVocabularyRegistry
        from zope.schema.vocabulary import SimpleVocabulary

        def vocabFactory(obj):
            return SimpleVocabulary.fromValues(["one", "two"])

        reg = getVocabularyRegistry()
        reg.register("my.vocab", vocabFactory)

        class T:
            f = None

        f = Choice(__name__="f", title="Test", vocabulary="my.vocab")
        t = T()

        # Bug: this would give "AttributeError: can't set attribute" on
        # clone.vocabulary
        f.bind(t)

    def test_fieldref_interfaces(self):
        from plone.registry import field
        from plone.registry import FieldRef
        from plone.registry.interfaces import IFieldRef
        from zope.schema.interfaces import ICollection

        listField = field.List(value_type=field.ASCIILine())
        ref = FieldRef("some.record", listField)

        self.assertTrue(ICollection.providedBy(ref))
        self.assertTrue(IFieldRef.providedBy(ref))


class TestMigration(unittest.TestCase):
    def setUp(self):
        setUp(self)

    def tearDown(self):
        testing.tearDown(self)

    def test_auto_migration(self):
        from BTrees.OOBTree import OOBTree
        from plone.registry import field
        from plone.registry.record import Record
        from plone.registry.registry import _Records
        from plone.registry.registry import Records
        from plone.registry.registry import Registry

        # Create an "old-looking registry"

        registry = Registry()
        registry._records = Records(registry)
        registry._records.data = OOBTree()

        f = field.TextLine(title="Foo")

        record = Record(f, "Bar")
        record.__dict__["field"] = f
        record.__dict__["value"] = "Bar"

        registry._records.data["foo.bar"] = record

        # Attempt to access it

        value = registry["foo.bar"]

        # Migration should have happened

        self.assertEqual(value, "Bar")
        self.assertEqual(registry.records["foo.bar"].field.title, "Foo")
        self.assertEqual(registry.records["foo.bar"].value, "Bar")

        self.assertFalse(isinstance(registry._records, Records))
        self.assertTrue(isinstance(registry._records, _Records))


def test_suite():
    return unittest.TestSuite(
        [
            doctest.DocFileSuite(
                "registry.rst",
                package="plone.registry",
                optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
                setUp=setUp,
                tearDown=testing.tearDown,
                checker=PolyglotOutputChecker(),
            ),
            doctest.DocFileSuite(
                "events.rst",
                package="plone.registry",
                optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
                setUp=setUp,
                tearDown=testing.tearDown,
                checker=PolyglotOutputChecker(),
            ),
            doctest.DocFileSuite(
                "field.rst",
                package="plone.registry",
                optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
                setUp=setUp,
                tearDown=testing.tearDown,
                checker=PolyglotOutputChecker(),
            ),
            unittest.defaultTestLoader.loadTestsFromTestCase(TestBugs),
            unittest.defaultTestLoader.loadTestsFromTestCase(TestMigration),
        ]
    )



from collections.abc import MutableMapping
from plone.registry.interfaces import IRecordsProxy
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.schema import getFieldsInOrder
from zope.schema.interfaces import RequiredMissing

import re


_marker = object()


@implementer(IRecordsProxy)
class RecordsProxy:
    """A proxy that maps an interface to a number of records"""

    def __init__(self, registry, schema, omitted=(), prefix=None):
        if prefix is None:
            prefix = schema.__identifier__ + "."
        elif not prefix.endswith("."):
            prefix += "."

        # skip __setattr__
        self.__dict__["__schema__"] = schema
        self.__dict__["__registry__"] = registry
        self.__dict__["__omitted__"] = omitted
        self.__dict__["__prefix__"] = prefix
        self.__dict__["__parent__"] = registry

        alsoProvides(self, schema)

    def __getattr__(self, name):
        if not self.__dict__ or name in self.__dict__.keys():
            return super().__getattr__(name)
        if name not in self.__schema__:
            raise AttributeError(name)
        value = self.__registry__.get(self.__prefix__ + name, _marker)
        if value is _marker:
            value = self.__schema__[name].missing_value
        return value

    def __setattr__(self, name, value):
        if name in self.__schema__:
            full_name = self.__prefix__ + name
            if full_name not in self.__registry__:
                raise AttributeError(name)
            self.__registry__[full_name] = value
        else:
            self.__dict__[name] = value

    def __repr__(self):
        return "<{} for {}>".format(
            self.__class__.__name__, self.__schema__.__identifier__
        )


class RecordsProxyCollection(MutableMapping):
    """A proxy that maps a collection of RecordsProxy objects"""

    _validkey = re.compile(r"([a-zA-Z][a-zA-Z0-9_.-]*)$").match

    # ord('.') == ord('/') - 1

    def __init__(
        self, registry, schema, check=True, omitted=(), prefix=None, factory=None
    ):
        if prefix is None:
            prefix = schema.__identifier__

        if not prefix.endswith("/"):
            prefix += "/"

        self.registry = registry
        self.schema = schema
        self.check = check
        self.omitted = omitted
        self.prefix = prefix
        self.factory = factory

    def __getitem__(self, key):
        if key in iter(self):
            prefix = self.prefix + key
            proxy = self.registry.forInterface(
                self.schema, self.check, self.omitted, prefix, self.factory
            )
            return proxy
        raise KeyError(key)

    def __iter__(self):
        min = self.prefix
        max = self.prefix[:-1] + "0"
        keys = self.registry.records.keys(min, max)
        len_prefix = len(self.prefix)
        last = None
        for name in keys:
            name = name[len_prefix:]
            if "." not in name:
                yield name
            else:
                key = name.rsplit(".", 1)[0]
                if key != last:
                    yield key
                    last = key

    def __len__(self):
        return len(tuple(iter(self)))

    def keys(self):
        return list(iter(self))

    def _validate(self, key):
        if not isinstance(key, str) or not self._validkey(key):
            raise TypeError(
                "expected a valid key (alphanumeric or underscore, starting "
                "with alpha)"
            )
        return str(key)

    def has_key(self, key):
        key = self._validate(key)
        prefix = self.prefix + key
        names = self.registry.records.keys(prefix + ".", prefix + "/")
        return bool(names)

    def add(self, key):
        key = self._validate(key)
        prefix = self.prefix + key
        self.registry.registerInterface(self.schema, self.omitted, prefix)
        proxy = self.registry.forInterface(self.schema, False, self.omitted, prefix)
        return proxy

    def __setitem__(self, key, value):
        key = self._validate(key)
        data = {}
        for name, field in getFieldsInOrder(self.schema):
            if name in self.omitted or field.readonly:
                continue
            attr = getattr(value, name, _marker)
            if attr is not _marker:
                data[name] = attr
            elif field.required and self.check:
                raise RequiredMissing(name)

        proxy = self.add(key)
        for name, attr in data.items():
            setattr(proxy, name, attr)

    def setdefault(self, key, failobj=None):
        if key not in self:
            if failobj is None:
                self.add(key)
            else:
                self[key] = failobj
        return self[key]

    def __delitem__(self, key):
        if key not in self:
            raise KeyError(key)
        prefix = self.prefix + key
        names = list(self.registry.records.keys(prefix + ".", prefix + "/"))
        for name in names:
            del self.registry.records[name]



from BTrees.OOBTree import OOBTree
from persistent import Persistent
from plone.registry.events import RecordAddedEvent
from plone.registry.events import RecordRemovedEvent
from plone.registry.fieldref import FieldRef
from plone.registry.interfaces import IFieldRef
from plone.registry.interfaces import InvalidRegistryKey
from plone.registry.interfaces import IPersistentField
from plone.registry.interfaces import IRecord
from plone.registry.interfaces import IRegistry
from plone.registry.record import Record
from plone.registry.recordsproxy import RecordsProxy
from plone.registry.recordsproxy import RecordsProxyCollection
from zope.component import queryAdapter
from zope.event import notify
from zope.interface import implementer
from zope.schema import getFieldNames
from zope.schema import getFieldsInOrder

import re
import warnings


@implementer(IRegistry)
class Registry(Persistent):
    """The persistent registry"""

    def __init__(self):
        self._records = _Records(self)

    # Basic value access API

    def __getitem__(self, name):
        # Fetch straight from records._values to avoid loading the field
        # as a separate persistent object
        return self.records._values[name]

    def get(self, name, default=None):
        # Fetch straight from records._values to avoid loading the field
        # as a separate persistent object
        return self.records._values.get(name, default)

    def __setitem__(self, name, value):
        # make sure we get the Record class' validation
        self.records[name].value = value

    def __contains__(self, name):
        return name in self.records._values

    # Records - make this a property so that it's readonly

    @property
    def records(self):
        # XXX: On-the-fly migration
        if isinstance(self._records, Records):
            self._migrateRecords()
        return self._records

    # Schema interface API

    def forInterface(self, interface, check=True, omit=(), prefix=None, factory=None):
        if prefix is None:
            prefix = interface.__identifier__

        if not prefix.endswith("."):
            prefix += "."

        if check:
            for name in getFieldNames(interface):
                if name not in omit and prefix + name not in self:
                    raise KeyError(
                        "Interface `{}` defines a field `{}`, for which "
                        "there is no record.".format(interface.__identifier__, name)
                    )

        if factory is None:
            factory = RecordsProxy

        return factory(self, interface, omitted=omit, prefix=prefix)

    def registerInterface(self, interface, omit=(), prefix=None):
        if prefix is None:
            prefix = interface.__identifier__

        if not prefix.endswith("."):
            prefix += "."

        for name, field in getFieldsInOrder(interface):
            if name in omit or field.readonly:
                continue
            record_name = prefix + name
            persistent_field = queryAdapter(field, IPersistentField)
            if persistent_field is None:
                raise TypeError(
                    "There is no persistent field equivalent for the field "
                    "`{}` of type `{}`.".format(name, field.__class__.__name__)
                )

            persistent_field.interfaceName = interface.__identifier__
            persistent_field.fieldName = name

            value = persistent_field.default

            # Attempt to retain the existing value
            if record_name in self.records:
                existing_record = self.records[record_name]
                value = existing_record.value
                bound_field = persistent_field.bind(existing_record)
                try:
                    bound_field.validate(value)
                except Exception:
                    value = persistent_field.default

            self.records[record_name] = Record(persistent_field, value, _validate=False)

    def collectionOfInterface(
        self, interface, check=True, omit=(), prefix=None, factory=None
    ):
        return RecordsProxyCollection(self, interface, check, omit, prefix, factory)

    # BBB

    def _migrateRecords(self):
        """BBB: Migrate from the old Records structure to the new. This is
        done the first time the "old" structure is accessed.
        """
        records = _Records(self)

        oldData = getattr(self._records, "data", None)
        if oldData is not None:
            for name, oldRecord in oldData.iteritems():
                oldRecord._p_activate()
                if "field" in oldRecord.__dict__ and "value" in oldRecord.__dict__:
                    records._fields[name] = oldRecord.__dict__["field"]
                    records._values[name] = oldRecord.__dict__["value"]

        self._records = records


class _Records:
    """The records stored in the registry. This implements dict-like access
    to records, where as the Registry object implements dict-like read-only
    access to values.
    """

    __parent__ = None

    # Similar to zope.schema._field._isdotted, but allows up to one '/'
    _validkey = re.compile(
        r"([a-zA-Z][a-zA-Z0-9_-]*)((?:\.[a-zA-Z0-9][a-zA-Z0-9_-]*)*)"
        r"([/][a-zA-Z0-9][a-zA-Z0-9_-]*)?((?:\.[a-zA-Z0-9][a-zA-Z0-9_-]*)*)$"
    ).match

    def __init__(self, parent):
        self.__parent__ = parent

        self._fields = OOBTree()
        self._values = OOBTree()

    def __setitem__(self, name, record):
        if not self._validkey(name):
            raise InvalidRegistryKey(record)
        if not IRecord.providedBy(record):
            raise ValueError("Value must be a record")

        self._setField(name, record.field)
        self._values[name] = record.value

        record.__name__ = name
        record.__parent__ = self.__parent__

        notify(RecordAddedEvent(record))

    def __delitem__(self, name):
        record = self[name]

        # unbind the record so that it won't attempt to look up values from
        # the registry anymore
        record.__parent__ = None

        del self._fields[name]
        del self._values[name]

        notify(RecordRemovedEvent(record))

    def __getitem__(self, name):
        field = self._getField(name)
        value = self._values[name]

        record = Record(field, value, _validate=False)
        record.__name__ = name
        record.__parent__ = self.__parent__

        return record

    def get(self, name, default=None):
        try:
            return self[name]
        except KeyError:
            return default

    def __nonzero__(self):
        return self._values.__nonzero__()

    def __len__(self):
        return self._values.__len__()

    def __iter__(self):
        return self._values.__iter__()

    def has_key(self, name):
        return self._values.__contains__(name)

    def __contains__(self, name):
        return self._values.__contains__(name)

    def keys(self, min=None, max=None):
        return self._values.keys(min, max)

    def maxKey(self, key=None):
        return self._values.maxKey(key)

    def minKey(self, key=None):
        return self._values.minKey(key)

    def values(self, min=None, max=None):
        return [self[name] for name in self.keys(min, max)]

    def items(self, min=None, max=None):
        return [(name, self[name]) for name in self.keys(min, max)]

    def setdefault(self, key, value):
        if key not in self:
            self[key] = value
        return self[key]

    def clear(self):
        self._fields.clear()
        self._values.clear()

    # Helper methods

    def _getField(self, name):
        field = self._fields[name]

        # Handle field reference pointers
        if isinstance(field, str):
            recordName = field
            while isinstance(field, str):
                recordName = field
                field = self._fields[recordName]
            field = FieldRef(recordName, field)

        return field

    def _setField(self, name, field):
        if not IPersistentField.providedBy(field):
            raise ValueError("The record's field must be an IPersistentField.")
        if IFieldRef.providedBy(field):
            if field.recordName not in self._fields:
                raise ValueError("Field reference points to non-existent record")
            self._fields[name] = field.recordName  # a pointer, of sorts
        else:
            field.__name__ = "value"
            self._fields[name] = field


class Records(_Records, Persistent):
    """BBB: This used to be the class for the _records attribute of the
    registry. Having this be a Persistent object was always a bad idea. We
    keep it here so that we can migrate to the new structure, but it should
    no longer be used.
    """

    def __init__(self, parent):
        warnings.warn(
            "The Records persistent class is deprecated and should not be " "used.",
            DeprecationWarning,
        )
        super().__init__(parent)



from setuptools import find_packages
from setuptools import setup

import os


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


version = "2.0.0"

description = "Registry for application settings (like debconf/ about:config)"
long_description = (
    read("README.rst")
    + "\n"
    + read("plone", "registry", "registry.rst")
    + "\n"
    + read("plone", "registry", "events.rst")
    + "\n"
    + read("plone", "registry", "field.rst")
    + "\n"
    + read("CHANGES.rst")
    + "\n"
)

setup(
    name="plone.registry",
    version=version,
    description=description,
    long_description=long_description,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="configuration registry",
    author="Martin Aspeli, Wichert Akkerman, Hanno Schlichting",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://pypi.org/project/plone.registry",
    license="GPL",
    packages=find_packages(),
    namespace_packages=["plone"],
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.8",
    install_requires=[
        "setuptools",
        "zope.component",
        "zope.dottedname",
        "zope.event",
        "zope.interface",
        "zope.schema",
    ],
    extras_require={"test": ["BTrees", "plone.schema"]},
    entry_points="""
    # -*- Entry points: -*-
    """,
)



from plone.registry.interfaces import IInterfaceAwareRecord
from plone.registry.interfaces import IRecordAddedEvent
from plone.registry.interfaces import IRecordEvent
from plone.registry.interfaces import IRecordModifiedEvent
from plone.registry.interfaces import IRecordRemovedEvent
from plone.registry.recordsproxy import RecordsProxy
from zope.component import adapter
from zope.component import subscribers
from zope.interface import implementer


@implementer(IRecordEvent)
class RecordEvent:
    def __init__(self, record):
        self.record = record

    def __repr__(self):
        return f"<{self.__class__.__name__} for {self.record.__name__}>"


@implementer(IRecordAddedEvent)
class RecordAddedEvent(RecordEvent):
    """record added"""


@implementer(IRecordRemovedEvent)
class RecordRemovedEvent(RecordEvent):
    """record removed"""


@implementer(IRecordModifiedEvent)
class RecordModifiedEvent(RecordEvent):
    def __init__(self, record, oldValue, newValue):
        super().__init__(record)
        self.oldValue = oldValue
        self.newValue = newValue


@adapter(IRecordEvent)
def redispatchInterfaceAwareRecordEvents(event):
    """When an interface-aware record received a record event,
    redispatch the event in a similar manner to the IObjectEvent redispatcher.

    Note that this means one IRecordModifiedEvent will be fired for each
    change to a record.
    """

    record = event.record

    if not IInterfaceAwareRecord.providedBy(record):
        return

    schema = record.interface
    if schema is None:
        return

    proxy = RecordsProxy(record.__parent__, schema)

    adapters = subscribers((proxy, event), None)
    for ad in adapters:
        pass  # getting them does the work
